package main

import (
	"flag"
	"fmt"
	"log"
	"net"
	"net/rpc"
	"time"

	drbg "github.com/davidlazar/go-drbg/hmac"

	. "dcnet"
)

type Server struct {
	addr       string
	numClients int
	ciphertext [][]byte
	responses  [][]byte

	ctxtChan      chan ClientCiphertext
	stxtChan      chan [][]byte
	waitResponses chan bool

	controlServer *rpc.Client
}

func NewServer(addr string) *Server {
	s := Server{
		addr:          addr,
		numClients:    0,
		ctxtChan:      make(chan ClientCiphertext),
		stxtChan:      make(chan [][]byte),
		waitResponses: make(chan bool),
	}

	cs, err := rpc.Dial("tcp", ControlServer)
	if err != nil {
		log.Fatal("dialing: ", err)
	}

	s.controlServer = cs
	return &s
}

func (s *Server) SendMessage(ct ClientCiphertext, _ *int) error {
	s.ctxtChan <- ct
	return nil
}

func (s *Server) GetCiphertext(_ int, stxt *[][]byte) error {
	*stxt = <-s.stxtChan
	return nil
}

func (s *Server) gatherClientCiphertexts() {
	s.ciphertext = make([][]byte, s.numClients)
	for i := range s.ciphertext {
		s.ciphertext[i] = make([]byte, MsgLen)
	}

	for m := 0; m < s.numClients; m++ {
		ct := <-s.ctxtChan

		rng := drbg.New(ct.SecretSeed)

		for i, ctxt := range ct.Ciphertext {
			Xor(ctxt, s.ciphertext[i])
			Xor(rng.Generate(MsgLen), s.ciphertext[i])
		}
	}
}

func (s *Server) sendServerCiphertext() {
	err := s.controlServer.Call("Server.AddServerCiphertext", s.ciphertext, nil)
	if err != nil {
		log.Fatal("client error: ", err)
	}
}

func (s *Server) getResponses() {
	var responses [][]byte
	err := s.controlServer.Call("Server.GetResponses", 0, &responses)
	if err != nil {
		log.Fatal("client error: ", err)
	}

	s.responses = responses
}

func (s *Server) XorResponses(mask []byte, r *[]byte) error {
	<-s.waitResponses

	response := make([]byte, RspLen)
	i := 0
L:
	for _, b := range mask {
		for j := 0; j < 8; j++ {
			if b&1 == 1 {
				Xor(s.responses[i], response)
			}
			b >>= 1
			i++
			if i >= s.numClients {
				break L
			}
		}
	}
	*r = response
	return nil
}

func (s *Server) round() {
	var r ServerRegistration
	err := s.controlServer.Call("Server.RegisterServer", s.addr, &r)
	if err != nil {
		log.Fatal("client error: ", err)
	}
	s.numClients = r.NumClients

	s.gatherClientCiphertexts()
	s.sendServerCiphertext()
	s.getResponses()
	close(s.waitResponses)
}

// TODO need a more principled approach to server reinitialization
// one approach is to work together with NewServer()
func (s *Server) cleanup() {
	s.numClients = 0
	s.waitResponses = make(chan bool)
}

func (s *Server) serverLoop() {
	for {
		s.round()

		time.Sleep(500 * time.Millisecond)
		s.cleanup()
	}
}

func main() {
	var addr *string = flag.String("l", ":50001", "listen addr:port")
	flag.Parse()

	s := NewServer(*addr)
	server := rpc.NewServer()
	server.Register(s)

	go s.serverLoop()

	l, e := net.Listen("tcp", *addr)
	if e != nil {
		log.Fatal("listen error:", e)
	}

	fmt.Println("anonymity server listening...")
	server.Accept(l)
}
