package main

import (
	"bufio"
	"bytes"
	"crypto/rand"
	"flag"
	"fmt"
	"log"
	mrand "math/rand"
	"net/rpc"
	"os"
	"time"

	drbg "github.com/davidlazar/go-drbg/hmac"

	. "dcnet"
)

type Client struct {
	anonymityServers []*rpc.Client
	controlServer    *rpc.Client

	numClients int
	clientSlot int

	messageChan chan []byte
	mode string
	rpcClient *rpc.Client
}

func NewClient() *Client {
	c := Client{
		anonymityServers: make([]*rpc.Client, 0, 32),
		messageChan:      make(chan []byte),
	}

	cs, err := rpc.Dial("tcp", ControlServer)
	if err != nil {
		log.Fatal("dialing: ", err)
	}

	c.controlServer = cs
	return &c
}

func (c *Client) sendMessage(message []byte) {
	seeds := make([][]byte, len(c.anonymityServers))
	prngs := make([]*drbg.DRBG, len(c.anonymityServers))
	for i := range seeds {
		seeds[i] = make([]byte, 32)
		rand.Read(seeds[i]) // TODO check error

		prngs[i] = drbg.New(seeds[i])
	}

	ctxt := make([][]byte, c.numClients)
	for n := 0; n < c.numClients; n++ {
		ctxt[n] = make([]byte, MsgLen)
		for _, rng := range prngs {
			s := rng.Generate(MsgLen)
			Xor(s, ctxt[n])
		}
	}

	Xor(message, ctxt[c.clientSlot])

	for i, server := range c.anonymityServers {
		ct := ClientCiphertext{seeds[i], ctxt}
		err := server.Call("Server.SendMessage", ct, nil)
		if err != nil {
			log.Fatal("client error: ", err)
		}
	}
}

func (c *Client) getResponse() []byte {
	masks := make([][]byte, len(c.anonymityServers))

	// number of bytes in each mask
	b := (c.numClients + 7) / 8
	for i := 1; i < len(c.anonymityServers); i++ {
		masks[i] = make([]byte, b)
		rand.Read(masks[i])
	}

	// generate mask
	mymask := make([]byte, b)
	SetBit(uint(c.clientSlot), true, mymask)
	masks[0] = mymask

	sum := Xors(masks)
	// xor all of the together to get the original one
	masks[0] = sum

	// TODO (Alex): We shouldn't need to randomly permute the servers, right?
	perm := mrand.Perm(len(c.anonymityServers))

	// xor and get the right response
	var responseXor []byte
	response := make([]byte, RspLen)
	for i, x := range perm {
		err := c.anonymityServers[i].Call("Server.XorResponses", masks[x], &responseXor)
		if err != nil {
			log.Fatal("client error: ", err)
		}

		Xor(responseXor, response)
	}

	return response
}

func (c *Client) round() {
	var r ClientRegistration
	err := c.controlServer.Call("Server.RegisterClient", 0, &r)
	if err != nil {
		log.Fatal("client error: ", err)
	}

	c.numClients = r.NumClients
	c.clientSlot = r.ClientSlot

	for _, server := range r.Servers {
		s, err := rpc.Dial("tcp", server)
		if err != nil {
			log.Fatal("dialing: ", err)
		}
		c.anonymityServers = append(c.anonymityServers, s)
	}

	var msg []byte
	select {
	case msg = <-c.messageChan:
		c.sendMessage(msg)
	default:
		c.sendMessage([]byte{})
	}

	response := c.getResponse()

	if msg != nil {
		if c.mode == "cmd" {
			fmt.Println(CString(response))
		} else if c.mode == "proxy" {
			err = c.rpcClient.Call("ProxyRPC.PutResponse", response, nil)
			if err != nil {
				log.Fatal("Failed at sending back response: ", err)
			}
		}
	}
}

func (c *Client) stdinReader() {
	scanner := bufio.NewScanner(os.Stdin)
	for scanner.Scan() {
		c.messageChan <- bytes.TrimSpace(scanner.Bytes())
	}

	if err := scanner.Err(); err != nil {
		fmt.Fprintln(os.Stderr, "reading standard input: ", err)
	}
}

func (c *Client) proxyReader() {
	for {
		req := make([]byte, MsgLen)
		err := c.rpcClient.Call("ProxyRPC.GetRequest", 0, &req)

		if err != nil {
			fmt.Fprintln(os.Stderr, "Could not get http request: ", err)
		}

		println(string(req))
		c.messageChan <- bytes.TrimSpace(req)
	}
}

func (c *Client) cleanup() {
	c.numClients = 0
	c.clientSlot = 0

	for _, server := range c.anonymityServers {
		server.Close()
	}
	c.anonymityServers = c.anonymityServers[:0]
}

func (c *Client) clientLoop() {
	for {
		c.round()
		c.cleanup()
		time.Sleep(500 * time.Millisecond)
	}
}

func main() {
	var mode *string = flag.String("m", "cmd", "mode [cmd|proxy]")
	flag.Parse()

	c := NewClient()
	c.mode = *mode

	if c.mode == "cmd" {
		go c.stdinReader()
	} else if c.mode == "proxy" {
		server, err := rpc.Dial("tcp", ":43330")
		if err != nil {
			log.Fatal("Cannot open connection:", err)
		}
		c.rpcClient = server
		go c.proxyReader()
	} else {
		fmt.Println("Bad mode argument")
		os.Exit(1)
	}
	c.clientLoop()
}
